import os
import angr
import logging

l = logging.getLogger("archr.analyzers.angr_state")

from . import Analyzer


class SimArchrMount(angr.state_plugins.filesystem.SimConcreteFilesystem):
    def __init__(self, target=None, **kwargs):
        super().__init__(**kwargs)
        self.target = target

    @angr.SimStatePlugin.memo
    def copy(self, memo):
        o = super().copy(memo)
        o.target = self.target
        return o

    def _load_file(self, guest_path):
        from docker.errors import NotFound  # pylint:disable=import-outside-toplevel
        try:
            target_path = self.target.realpath(guest_path)
            content = self.target.retrieve_contents(target_path)
            return angr.SimFile(name='file://' + target_path, content=content, size=len(content))
        except NotFound:
            return None

    def _get_stat(self, guest_path, dereference=False):
        if dereference:
            guest_path = self.target.realpath(guest_path)
        stat_output = self.target.run_command([
            "stat", "-c", "%n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %W %o %B",
            guest_path
        ]).communicate()[0].decode().split()

        # parse output
        # st_mode, device number, major type, minor type are in hex
        for i in [3, 6, 9, 10]:
            stat_output[i] = int(stat_output[i], 16)
        stat_output[1:] = [int(x) for x in stat_output[1:]] # others are int

        # %n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %W %o %C
        (
            _, # %n file name is ignored
            st_size, # %s total size, in bytes
            st_blocks, # %b number of blocks allocated (see %B)
            st_mode, # %f raw mode in hex
            st_uid, # %u     user ID of owner
            st_gid, # %g     group ID of owner
            st_dev, # %D     device number in hex
            st_ino, # %i     inode number
            st_nlink, # %h number of hard links
            _, # %t major device type in hex, for character/block device special files
            _, # %T minor device type in hex, for character/block device special files
            st_atime, # %X time of last access, seconds since Epoch
            st_mtime, # %Y time of last data modification, seconds since Epoch
            st_ctime, # %Z time of last status change, seconds since Epoch
            _, # %W time of file birth, seconds since Epoch; 0 if unknown
            _, # %o optimal I/O transfer size hint
            st_blksize, # %B block size
        ) = stat_output
        st_mtime_ns = st_mtime * 1000000
        st_atime_ns = st_atime * 1000000
        st_ctime_ns = st_ctime * 1000000
        WRONG_rdev = 0 # TODO: build from major and minor: https://code.woboq.org/qt5/include/bits/sysmacros.h.html
        st = angr.state_plugins.filesystem.Stat(
            st_dev, st_ino, st_nlink, st_mode, st_uid, st_gid,
            WRONG_rdev, st_size, st_blksize, st_blocks,
            st_atime, st_atime_ns, st_mtime, st_mtime_ns, st_ctime, st_ctime_ns
        )
        return st

class SimArchrProcMount(SimArchrMount):
    def _load_file(self, guest_path):
        '''
        For some reason the procfs files cannot be fetched with the docker API, so here we just `cat` instead
        :param guest_path:
        :return:
        '''
        # attempt normal loading first
        file = super()._load_file(os.path.join('/proc', guest_path))
        if file is not None:
            return file

        target_path = self.target.realpath(os.path.join('/proc', guest_path.lstrip(os.path.sep)))
        content, error = self.target.run_command(["cat", target_path]).communicate()
        if not error:
            return angr.SimFile(name='file://' + target_path, content=content, size=len(content))
        else:
            assert ': No such file or directory' in error
            return None

    def _get_stat(self, guest_path, dereference=False):
        return super()._get_stat(os.path.join('/proc', guest_path.lstrip(os.path.sep)), dereference=dereference)


class angrStateAnalyzer(Analyzer):
    """
    Constructs an angr state (full init variety) to match the target precisely
    """

    def __init__(self, target, project_analyzer):
        super().__init__(target)
        self.project_analyzer = project_analyzer

    def fire(self, **kwargs): #pylint:disable=arguments-differ
        project = self.project_analyzer.fire()
        if 'cwd' not in kwargs:
            cwd = os.path.dirname(self.project_analyzer.target.target_path)
            kwargs['cwd'] = bytes(cwd, 'utf-8')

        concrete_fs = kwargs.pop('concrete_fs', True)
        chroot = kwargs.pop('chroot', '/ARCHR-INVALID')
        stack_end = kwargs.pop('stack_end', self.project_analyzer._mem_mapping.get('[stack-end]', None))
        args = kwargs.pop('args', self.target.main_binary_args)
        env = kwargs.pop('env', self.target.target_env)
        brk = kwargs.pop('brk', self.project_analyzer._mem_mapping.get('[heap]', None))

        s = project.factory.full_init_state(
            concrete_fs=concrete_fs,
            chroot=chroot,
            stack_end=stack_end,
            args=args,
            env=env,
            brk=brk,
            **kwargs
        )
        s.fs.mount("/", SimArchrMount(self.target))
        s.fs.set_state(s)
        return s



import os
import logging

l = logging.getLogger("archr.analyzers.memory_map")

from . import ContextAnalyzer

class GDBServerAnalyzer(ContextAnalyzer):
    """
    Launches a gdb server.
    """

    REQUIRED_IMPLANT = "gdbserver"

    def fire_context(self, port=31337, aslr=False, **kwargs): #pylint:disable=arguments-differ
        fire_path = os.path.join(self.target.tmpwd, "gdbserver", "fire")
        return self.target.flight_context(args_prefix=[fire_path, "0.0.0.0:%d"%port], aslr=aslr, **kwargs)



import subprocess
from typing import TYPE_CHECKING
import logging

try:
    import angr
except ImportError:
    syscall_agent = None

try:
    import syscall_agent
except ImportError:
    syscall_agent = None


from . import Analyzer

if TYPE_CHECKING:
    from .angr_project import angrProjectAnalyzer


_l = logging.getLogger(__name__)


class angrUltimateTracerAnalyzer(Analyzer):
    """
    Construct an angr project with ultimate tracer enabled. All syscalls will be out-sourced to an external syscall
    agent.
    """
    def __init__(self, target, project_analyzer: 'angrProjectAnalyzer'):
        if angr is None or syscall_agent is None:
            raise ImportError("Failed to import angr or syscall_agent. Make sure angr and syscall_agent are installed.")

        super().__init__(target)
        self.project_analyzer: 'angrProjectAnalyzer' = project_analyzer

    @staticmethod
    def _invoke_syscall_agent(project: 'angr.Project') -> subprocess.Popen:
        """
        Invoke the expected syscall agent.
        """
        agent = syscall_agent.manager.get_agent(project.arch.name)
        if agent is None:
            raise RuntimeError("Cannot find a syscall agent for project %r (architecture %s)" % (project,
                                                                                                 project.arch.name))
        project.bureau.start()  # get it ready to receive connections
        proc = agent.launch("tcp://127.0.0.1:%d" % project.bureau.zmq_port)  # launch the agent process
        return proc

    def make_project(self):
        if self.project_analyzer.project is not None:
            return self.project_analyzer.project

        engine = angr.engines.UberEngineSyscallTracing
        if self.project_analyzer.project is not None:
            _l.warning("An angr project was created. Destroying it.")
            self.project_analyzer.project = None

        project = self.project_analyzer.fire(project_kwargs={
            'auto_load_libs': True,
            'engine': engine,
            'use_sim_procedures': False,
        })

        return project

    def fire(self, *args, state: 'angr.SimState'=None, **kwargs):#pylint:disable=arguments-differ
        project = self.make_project()
        proc = self._invoke_syscall_agent(project)

        if state is None:
            raise ValueError('"state" must be specified')

        sim_manager = project.factory.simulation_manager(state)
        sim_manager.explore()

        # terminate the agent
        proc.terminate()



import contextlib
import subprocess
import tempfile
import logging
import signal
import shutil
import glob
import re
import os

from io import BytesIO


l = logging.getLogger("archr.analyzers.qemu_tracer")

from . import ContextAnalyzer
from .. import _angr_available
if _angr_available:
    import angr

from ..utils import filter_strace_output, get_file_maps

class QEMUTracerError(BaseException):
    pass

class QemuTraceResult:
    # results
    returncode = None
    signal = None
    crashed = None
    timed_out = None

    # introspection
    trace = None
    mapped_files = None
    crash_address = None
    base_address = None
    image_base = None
    entry_point = None
    magic_contents = None
    halfway_core_path = None
    core_path = None
    taint_fd = None

    def tracer_technique(self, **kwargs):
        return angr.exploration_techniques.Tracer(self.trace, crash_addr=self.crash_address, **kwargs)

_trace_old_re = re.compile(br'Trace (.*) \[(?P<addr>.*)\].*')
_trace_new_re = re.compile(br'Trace (.*) \[(?P<something1>.*)\/(?P<addr>.*)\/(?P<flags>.*)\].*')

class QEMUTracerAnalyzer(ContextAnalyzer):
    REQUIRED_IMPLANT = "shellphish_qemu"

    def __init__(self, target, timeout=10, ld_linux=None, ld_preload=None, library_path=None, seed=None, **kwargs):
        super().__init__(target, **kwargs)
        self.timeout = timeout
        self.ld_linux = ld_linux
        self.ld_preload = ld_preload
        self.library_path = library_path
        self.seed = seed

    def pickup_env(self):
        for e in self.target.target_env:
            key, value = e.split('=', 1)
            if key == 'LD_PRELOAD' and self.ld_preload is None:
                self.ld_preload = value
            if key == 'LD_LIBRARY_PATH' and self.library_path is None:
                self.library_path = value

    @contextlib.contextmanager
    def _target_mk_tmpdir(self):
        tmpdir = tempfile.mktemp(prefix="/tmp/tracer_target_")
        self.target.run_command(["mkdir", tmpdir]).wait()
        self.target.run_command(["chmod", "777", tmpdir]).wait()
        try:
            yield tmpdir
        finally:
            self.target.run_command(["rm", "-rf", tmpdir])

    @staticmethod
    @contextlib.contextmanager
    def _local_mk_tmpdir():
        tmpdir = tempfile.mkdtemp(prefix="/tmp/tracer_local_")
        try:
            yield tmpdir
        finally:
            with contextlib.suppress(FileNotFoundError):
                shutil.rmtree(tmpdir)

    @staticmethod
    def line_iter(content):
        io = BytesIO(content)
        for line in io:
            yield line.strip()

    @contextlib.contextmanager
    def fire_context(self, record_trace=True, record_magic=False, save_core=False, record_file_maps=False, # pylint: disable=arguments-differ
                     crash_addr=None, trace_bb_addr=None, taint=None, **kwargs): # pylint:disable=arguments-differ
        with self._target_mk_tmpdir() as tmpdir:
            tmp_prefix = tempfile.mktemp(dir='/tmp', prefix="tracer-")
            target_trace_filename = tmp_prefix + ".trace" if record_trace else None
            target_magic_filename = tmp_prefix + ".magic" if record_magic else None
            local_core_filename = tmp_prefix + ".core" if save_core else None
            local_halfway_core_filename = tmp_prefix + f'.halfway_{hex(crash_addr[0])}_{crash_addr[1]}.core' if crash_addr else None

            target_cmd = self._build_command(
                trace_filename=target_trace_filename,
                magic_filename=target_magic_filename,
                coredump_dir=tmpdir,
                crash_addr=crash_addr,
                start_trace_addr=trace_bb_addr,
                taint=taint)
            
            l.debug("launch QEMU with command: %s", ' '.join(target_cmd))
            r = QemuTraceResult()

            try:
                with self.target.flight_context(target_cmd, timeout=self.timeout, result=r, **kwargs) as flight:
                    yield flight
            except subprocess.TimeoutExpired:
                r.timed_out = True
            else:
                r.timed_out = False
                r.returncode = flight.process.returncode

                # did a crash occur?
                if r.returncode in [ 139, -11 ]:
                    r.crashed = True
                    r.signal = signal.SIGSEGV
                elif r.returncode == [ 132, -9 ]:
                    r.crashed = True
                    r.signal = signal.SIGILL

            l.debug("Qemu tracer returned with code=%s timed_out=%s crashed=%s signal=%s",
                    r.returncode, r.timed_out, r.crashed, r.signal)

            if local_core_filename or crash_addr:
                # choose the correct core dump to retrieve
                with self._local_mk_tmpdir() as local_tmpdir:
                    self.target.retrieve_into(tmpdir, local_tmpdir)
                    target_cores = glob.glob(os.path.join(local_tmpdir, '*', 'qemu_*.core'))
                    tmp_crash_core_path = None
                    tmp_halfway_core_path = None

                    for x in target_cores:
                        if 'crash' in x.rsplit("_")[-1]:
                            tmp_crash_core_path = x
                        if 'coreaddr' in x.rsplit("_")[-1]:
                            tmp_halfway_core_path = x

                    if tmp_crash_core_path is None and len(target_cores) == 1:
                        tmp_crash_core_path = target_cores[0]

                    # sanity check core dumps
                    if save_core and not tmp_crash_core_path:
                        raise QEMUTracerError("the target didn't crash inside qemu or no corefile was created!" + 
                                              "Make sure you launch it correctly!\n" + 
                                              "command: %s" % ' '.join(target_cmd))
                    if crash_addr and not tmp_halfway_core_path:
                        raise QEMUTracerError("the target didn't generate a halfway core file!" +
                                         "command: %s" % ' '.join(target_cmd))

                    if local_core_filename and tmp_crash_core_path:
                        shutil.move(tmp_crash_core_path, local_core_filename)
                    if local_halfway_core_filename and tmp_halfway_core_path:
                        shutil.move(tmp_halfway_core_path, local_halfway_core_filename)
                    r.core_path = local_core_filename
                    r.halfway_core_path = local_halfway_core_filename

            if target_trace_filename:
                trace = self.target.retrieve_contents(target_trace_filename)
                trace_iter = self.line_iter(trace)

                # Find where qemu loaded the binary. Primarily for PIE
                try:
                    # the image base is the first mapped address in the page dump following the log line 'guest_base'
                    for t in trace_iter:
                        if t.startswith(b"guest_base"):
                            # iterate to the appropriate line
                            next(trace_iter)
                            next(trace_iter)
                            t = next(trace_iter)
                            # parse out the first line
                            r.image_base = int(t.split(b'-')[0],16)
                            break

                    r.base_address = int(next(t.split()[1] for t in trace_iter if t.startswith(b"start_code")), 16) #pylint:disable=stop-iteration-return

                    # for a dynamically linked binary, the entry point is in the runtime linker
                    # in this case it can be useful to keep track of the entry point
                    r.entry_point = int(next(t.split()[1] for t in trace_iter if t.startswith(b"entry")), 16)
                except StopIteration as e:
                    raise QEMUTracerError("The trace does not include any data. Did you forget to chmod +x the binary?") from e

                # record the trace
                _trace_re = _trace_old_re if self.target.target_os == 'cgc' else _trace_new_re
                r.trace = [
                    int(_trace_re.match(t).group('addr'), 16) for t in trace_iter if t.startswith(b"Trace ")
                ]

                endings = trace.rsplit(b'\n', 3)[1:3]

                if r.crashed:
                    # grab the taint_fd
                    if not endings[0].startswith(b"qemu: last read marker was read through fd:"):
                        l.error(
                            "Unexpected status line from qemu tracer. Cannot get the last read marker to set taint_fd. "
                            "Please make sure you are using the latest shellphish-qemu.")
                    else:
                        r.taint_fd = int(re.search(br'\[(\d+)\]', endings[0]).group(1))
                        l.debug("Detected the tainted fd to be %s", r.taint_fd)
                    # grab the faulting address
                    lastline = endings[-1]
                    if lastline.startswith(b"Trace") or lastline.find(b"Segmentation") == -1:
                        l.warning("Trace return code was less than zero, but the last line of the trace does not"
                                  "contain the uncaught exception error from qemu."
                                  "If using an older version of shellphish_qemu try using 'ulimit -Sc 0' or "
                                  "updating to a newer version of shellphish_qemu.")
                    r.crash_address = int(lastline.split(b'[')[1].split(b']')[0], 16)
                    l.debug("Detected the crashing address at %s", hex(r.crash_address))

                l.debug("Trace consists of %d basic blocks", len(r.trace))

                if record_file_maps:
                    strace_lines = filter_strace_output([line.decode('utf-8') for line in self.line_iter(trace)])
                    r.mapped_files = get_file_maps(strace_lines)

                # remove the trace file on the target
                self.target.remove_path(target_trace_filename)

            if target_magic_filename:
                r.magic_contents = self.target.retrieve_contents(target_magic_filename)
                if len(r.magic_contents) != 0x1000:
                    raise QEMUTracerError("Magic content read from QEMU improper size, should be a page in length")

                # remove the magic file on the target
                self.target.remove_path(target_magic_filename)

    @staticmethod
    def qemu_variant(target_os, target_arch, record_trace):
        """
        Need to know if we're tracking or not, specifically for what cgc qemu to use.
        """

        if target_os == 'cgc':
            suffix = "tracer" if record_trace else "base"
            qemu_variant = "shellphish-qemu-cgc-%s" % suffix
        else:
            qemu_variant = "shellphish-qemu-linux-%s" % target_arch

        return qemu_variant

    def _build_command(self, trace_filename=None, magic_filename=None, coredump_dir=None,
                       report_bad_args=None, crash_addr=None, start_trace_addr=None, taint=None):
        """
        Here, we build the tracing command.
        """

        #
        # First, the implant invocation
        #

        qemu_variant = self.qemu_variant(self.target.target_os, self.target.target_arch, trace_filename is not None)
        qemu_path = os.path.join(self.target.tmpwd, "shellphish_qemu", qemu_variant)
        fire_path = os.path.join(self.target.tmpwd, "shellphish_qemu", "fire")
        cmd_args = [fire_path, qemu_path]
        if coredump_dir:
            cmd_args += [ "-C", coredump_dir ]
        if crash_addr:
            cmd_args += [ "-A", '0x{:x}:{}'.format(*crash_addr) ]
        if start_trace_addr:
            cmd_args += [ "-T", '0x{:x}:{}'.format(*start_trace_addr) ]
        if taint:
            cmd_args += [ "-M", taint.hex()]

        #
        # Next, we build QEMU options.
        #

        # hardcode an argv[0]
        #cmd_args += [ "-0", program_args[0] ]

        # record trace
        if trace_filename:
            flags = "nochain,exec,page,strace" if 'cgc' not in qemu_variant else "exec"
            cmd_args += ["-d", flags, "-D", trace_filename]
        else:
            if 'cgc' in qemu_variant:
                cmd_args += ["-enable_double_empty_exiting"]

        # save CGC magic page
        if magic_filename:
            if 'cgc' not in qemu_variant:
                raise QEMUTracerError("Specified magic page dump on non-cgc architecture")
            cmd_args += ["-magicdump", magic_filename]

        if self.seed is not None:
            cmd_args.append("-seed")
            cmd_args.append(str(self.seed))

        if report_bad_args:
            cmd_args += ["-report_bad_args"]

        # Memory limit option is only available in shellphish-qemu-cgc-*
        if 'cgc' in qemu_variant:
            cmd_args += ["-m", "8G"]

        if 'cgc' not in qemu_variant and "LD_BIND_NOW=1" not in self.target.target_env:
            cmd_args += ['-E', 'LD_BIND_NOW=1']

        if self.ld_preload:
            cmd_args += ['-E', 'LD_PRELOAD=' + self.ld_preload]

        if self.library_path and not self.ld_linux:
            cmd_args += ['-E', 'LD_LIBRARY_PATH=' + self.library_path]

        # now set up the loader
        if self.ld_linux:
            cmd_args += [self.ld_linux]
            if self.library_path:
                cmd_args += ['--library-path', self.library_path]

        # Now, we add the program arguments.
        cmd_args += ["--"] # separate QEMU arguments and target arguments
        cmd_args += self.target.target_args

        return cmd_args



try:
    from setuptools import setup
    from setuptools import find_packages
    packages = find_packages()
except ImportError:
    from distutils.core import setup
    import os
    packages = [x.strip('./').replace('/','.') for x in os.popen('find -name "__init__.py" | xargs -n1 dirname').read().strip().split('\n')]

if bytes is str:
    raise Exception("This module is designed for python 3 only.")

setup(
    name='archr',
    version='9.0.10651',
    python_requires='>=3.6',
    packages=packages,
    package_data = {
        'archr': ['implants/*.sh', 'implants/*/*']
    },
    install_requires=[
        'shellphish_qemu>=0.12.2',
        'pygdbmi',
        'docker',
        'nclib>=1.0.0rc3',
        'patchelf-wrapper',
        'cle==9.0.10651',
        'ply'
    ],
    extras_require = {
        'angr':  ["angr==9.0.10651"],
        'qtrace': ["qtrace"],
    },
    description='Target-centric program analysis.',
    url='https://github.com/angr/archr',
    classifiers=["Operating System :: POSIX :: Linux"],
)



import contextlib
import logging
import os

from . import ContextAnalyzer

l = logging.getLogger("archr.analyzers.strace")

def super_yama():
    with open("/proc/sys/kernel/yama/ptrace_scope", 'rb') as c:
        if c.read().strip() != b"0":
            l.warning("/proc/sys/kernel/yama/ptrace_scope needs to be '0'. I am setting this system-wide.")
            os.system(_super_yama_cmd)

_super_yama_cmd = "echo 0 | docker run --rm --privileged -i ubuntu tee /proc/sys/kernel/yama/ptrace_scope"

class STraceAnalyzer(ContextAnalyzer):
    """
    Launches a process under strace
    """

    REQUIRED_BINARY = "/usr/bin/strace"

    @contextlib.contextmanager
    def fire_context(self, trace_args=None, args_prefix=None, **kwargs): #pylint:disable=arguments-differ
        """
        Starts strace with a fresh process.

        :param kwargs: Additional arguments to run_command
        :return: Target instance returned by run_command
        """

        fire_path = os.path.join(self.target.tmpwd, "strace", "fire")
        args_prefix = (args_prefix or []) + [fire_path] + (trace_args or []) + ["--"]
        with self.target.flight_context(args_prefix=args_prefix, **kwargs) as flight:
            yield flight
        try:
            flight.result = flight.process.stderr.read() # illegal, technically
        except ValueError:
            flight.result = b''


class STraceAttachAnalyzer(ContextAnalyzer):
    """
    Attaches to a process with strace
    """

    REQUIRED_BINARY = "/usr/bin/strace"

    @contextlib.contextmanager
    def fire_context(self, pid, trace_args=None, args_prefix=None, **kwargs): #pylint:disable=arguments-differ
        """
        Starts strace attaching to a given process

        :param pid: PID of target process, if already existing
        :param kwargs: Additional arguments to run_command
        :return: Target instance returned by run_command
        """

        super_yama()

        fire_path = os.path.join(self.target.tmpwd, "strace", "fire")
        args_prefix = (args_prefix or []) + [fire_path] + (trace_args or []) + ["-p", str(pid), "--"]
        with self.target.flight_context(args_prefix=args_prefix, **kwargs) as flight:
            yield flight
        flight.result = flight.process.stderr.read()
